from dataclasses import dataclass
from typing import Optional, Tuple

import torch

from torch_gauge.o3.functional import NormContraction1d, NormContraction2d


@dataclass
class SphericalTensor:
    r"""
    The SphericalTensor class tracks the SO(3) representation indices
     of a flattened data tensor for efficient group algebra operations.

    All angular and magnetic quantum number indices are treated as equivalent,
     and all non-degenerate feature channels are regarded as the "principal"
     indices, n. Note that the angular quantum number indices (m) for each l
     range from -l to l as in the physicist's convention.

    Two indexing tensors are maintained, ``metadata`` and ``rep_layout``.
    Attributes:
        data_ten (torch.Tensor): the tensor storing auto-differentiable data.
        metadata (torch.LongTensor): the minimal information which specifies
            the number of non-degenerate feature channels for each dimension
            and each l (and the associated m(s)).
        rep_layout (Tuple[torch.LongTensor]): stores the pointer (l[i], m[i], n[i])
            for each index i along the SO(3) representation dimension of the
            flattened data tensor.
    """

    def __init__(
        self,
        data_ten: torch.Tensor,
        rep_dims: Tuple[int, ...],
        metadata: torch.LongTensor,
        rep_layout: Optional[Tuple[torch.LongTensor, ...]] = None,
        num_channels: Optional[Tuple[int, ...]] = None,
    ):
        """
        Args:
            data_ten: the underlying flattened data tensor in the
             strictly incremental [l,m,n] order, for which the rep_dim has N elements
            rep_dims: the spherical tensor representation dimension(s). When
             n_rep_dims > 1, must be contiguous dimensions.
            metadata: (n_rep_dims x n_l) specification of the number of unique
             channels (n) for each angular momentum index l.
            rep_layout (optional): the (3xN) index tensor for (l[i],m[i],n[i]).
            num_channels (optional): the number of unique channels per rep_dim.
        """
        self.ten = data_ten
        self.metadata = metadata
        self.rep_dims = rep_dims
        self._norm_eps = 1e-4
        if rep_layout:
            self.rep_layout = rep_layout
        else:
            if len(rep_dims) > 1:
                assert rep_dims[1] - rep_dims[0] == 1
            self.rep_layout = self.generate_rep_layout()
        if num_channels:
            self.num_channels = num_channels
        else:
            self.num_channels = tuple(torch.sum(self.metadata, dim=1).long().tolist())

    @property
    def shape(self):
        return self.ten.shape

    def self_like(self, new_data_ten):
        """Returns a SphericalTensor with identical layout but new data tensor"""
        return SphericalTensor(
            new_data_ten,
            rep_dims=self.rep_dims,
            metadata=self.metadata,
            rep_layout=self.rep_layout,
            num_channels=self.num_channels,
        )

    def mul_(self, other: "SphericalTensor"):
        # Be careful that this operation will break equivariance
        self.ten.mul_(other.ten)
        return self.ten

    def add_(self, other: "SphericalTensor"):
        self.ten.add_(other.ten)
        return self.ten

    def __mul__(self, other: "SphericalTensor"):
        return self.self_like(self.ten * other.ten)

    def __add__(self, other: "SphericalTensor"):
        return self.self_like(self.ten + other.ten)

    def scalar_mul(self, other: torch.Tensor, inplace=False):
        """
        Performing representation-wise scalar multiplication.
        Args:
            other (torch.Tensor): along rep_dims, the shape must be the number
            of unique channels (n); along all other dimensions the shape must
            be the same as the data tensor.
            inplace (bool): If true, self.ten is updated in-place.
        """
        if len(self.rep_dims) == 1:
            broadcasted_other = torch.index_select(
                other, dim=self.rep_dims[0], index=self.rep_layout[0][2, :]
            )
        elif len(self.rep_dims) == 2:
            broadcasted_other = torch.index_select(
                other.view(
                    *other.shape[: self.rep_dims[0]],
                    self.num_channels[0] * self.num_channels[1],
                    *other.shape[self.rep_dims[1] + 1 :],
                ),
                dim=self.rep_dims[0],
                index=(
                    self.rep_layout[0][2, :].unsqueeze(1) * self.num_channels[1]
                    + self.rep_layout[1][2, :].unsqueeze(0)
                ).view(-1),
            ).view_as(self.ten)
        else:
            raise NotImplementedError

        if inplace:
            self.ten.mul_(broadcasted_other)
            return self
        else:
            return self.self_like(self.ten * broadcasted_other)

    def dot(self, other: "SphericalTensor", dim: int):
        """
        Performing inner product along a representation dimension.
        If self.n_rep_dim==1, a torch.Tensor is returned;
        if self.n_rep_dim==2, a SphericalTensor with n_rep_dim==1 is returned.
        Args:
            other (SphericalTensor): must be a 1-d spherical tensor
            with the same number of dimensions and broadcastable to self.ten.
            When self.n_rep_dim==2, the kept dimension in self.rep_dims
             of other must be singleton.
            dim (int): the dimension to perform the inner product.
        """
        dotdim_idx = self.rep_dims.index(dim)
        assert other.rep_dims[0] == dim
        assert torch.all(self.metadata[dotdim_idx].eq(other.metadata[0]))
        out_ten = self.ten.mul(other.ten).sum(dim)
        if len(self.rep_dims) == 1:
            return out_ten
        elif len(self.rep_dims) == 2:
            dimid_kept = 1 - dotdim_idx
            assert other.ten.shape[self.rep_dims[dimid_kept]] == 1
            return SphericalTensor(
                out_ten,
                rep_dims=(self.rep_dims[dimid_kept],),
                metadata=self.metadata[dimid_kept].unsqueeze(0),
                rep_layout=(self.rep_layout[dimid_kept],),
                num_channels=(self.num_channels[dimid_kept],),
            )
        else:
            raise NotImplementedError

    def rep_dot(self, other: "SphericalTensor", dim: int):
        """
        Performing channel-wise inner products.
        If self.n_rep_dim==1, a torch.Tensor is returned;
        if self.n_rep_dim==2, a SphericalTensor with n_rep_dim==1 is returned.
        Args:
            other (SphericalTensor): must be a 1-d spherical tensor
            with the same number of dimensions and broadcastable to self.ten.
            dim (int): the dimension to perform the channel-wise inner product.
        """
        dotdim_idx = self.rep_dims.index(dim)
        assert other.rep_dims[0] == dim
        assert torch.all(self.metadata[dotdim_idx].eq(other.metadata[0]))
        singleton_shape = tuple(
            self.ten.shape[d] if d == dim else 1 for d in range(self.ten.dim())
        )
        mul_ten = self.ten * other.ten
        out_shape = list(mul_ten.shape)
        out_shape[dim] = self.num_channels[dotdim_idx]
        out_ten = torch.zeros(
            out_shape, device=mul_ten.device, dtype=mul_ten.dtype
        ).scatter_add_(
            dim=dim,
            index=self.rep_layout[dotdim_idx][2, :]
            .view(*singleton_shape)
            .expand_as(mul_ten),
            src=mul_ten,
        )
        if len(self.rep_dims) == 1:
            return out_ten
        elif len(self.rep_dims) == 2:
            # A little trick here
            dimid_kept = 1 - dotdim_idx
            return SphericalTensor(
                out_ten,
                rep_dims=(self.rep_dims[dimid_kept],),
                metadata=self.metadata[dimid_kept].unsqueeze(0),
                rep_layout=(self.rep_layout[dimid_kept],),
                num_channels=(self.num_channels[dimid_kept],),
            )
        else:
            raise NotImplementedError

    def rep_outer(self, other: "SphericalTensor") -> "SphericalTensor":
        """
        Returns the batched outer product of two 1-d spherical tensors
        The rep_dim and metadata of self and other much be the same
        """
        assert len(self.rep_dims) == 1
        assert len(other.rep_dims) == 1
        assert (
            self.rep_dims[0] == other.rep_dims[0]
        ), "The representation dimensions of self and other must be contiguous"
        odim = self.rep_dims[0]
        out_ten = self.ten.unsqueeze(odim + 1).mul(other.ten.unsqueeze(odim))
        out_metadata = torch.cat([self.metadata, other.metadata], dim=0)
        out_rep_layout = (
            self.rep_layout[0],
            other.rep_layout[0],
        )
        return SphericalTensor(
            out_ten,
            rep_dims=(odim, odim + 1),
            metadata=out_metadata,
            rep_layout=out_rep_layout,
            num_channels=(self.num_channels, other.num_channels),
        )

    def fold(self, stride: int, update_self=False) -> "SphericalTensor":
        """
        Fold the chucked representation channels of a 1-d SphericalTensor to a new dimension
        """
        assert len(self.rep_dims) == 1
        assert torch.all(torch.fmod(self.metadata[0], stride) == 0), (
            f"The number of channels for theSphericalTensor to be folded must be multiples of "
            f"stride, got ({self.metadata}, {stride}) instead"
        )
        new_ten = self.ten.unflatten(
            dim=self.rep_dims[0],
            sizes=(self.shape[self.rep_dims[0]] // stride, stride),
        )
        new_metadata = self.metadata // stride
        new_rep_layout = (self.rep_layout[0][:, ::stride],)
        new_num_channels = (self.num_channels[0] // stride,)
        if update_self:
            self.ten = new_ten
            self.metadata = new_metadata
            self.rep_layout = new_rep_layout
            self.num_channels = new_num_channels
            return self
        else:
            return SphericalTensor(
                new_ten,
                rep_dims=self.rep_dims,
                metadata=new_metadata,
                rep_layout=new_rep_layout,
                num_channels=new_num_channels,
            )

    def transpose_repdims(self, inplace=False):
        assert (
            len(self.rep_dims) == 2
        ), "transpose_repdims only supports 2d SphericalTensor"
        ten_t = torch.transpose(self.ten, *self.rep_dims).contiguous()
        dims_t = self.rep_dims
        metadata_t = self.metadata[(1, 0), :]
        rep_layout_t = self.rep_layout[::-1]
        num_channels_t = self.num_channels[::-1]
        if inplace:
            self.ten = ten_t
            self.rep_dims = dims_t
            self.metadata = metadata_t
            self.rep_layout = rep_layout_t
            self.num_channels = num_channels_t
            return self
        else:
            return SphericalTensor(
                ten_t,
                rep_dims=dims_t,
                metadata=metadata_t,
                rep_layout=rep_layout_t,
                num_channels=num_channels_t,
            )

    def invariant(self) -> torch.Tensor:
        """
        Returns the invariant content
        When self.n_rep_dim==1, the l=0 channels are retained;
        When self.n_rep_dim==2, the (l1=0, l2=0) channels are also contracted.
        """
        if len(self.rep_dims) == 1:
            l0_length = self.metadata[0, 0]
            ops_dim = self.rep_dims[0]
            data_l0 = torch.narrow(self.ten, dim=ops_dim, start=0, length=l0_length)
            norm_shape = list(self.shape)
            norm_shape[ops_dim] = self.num_channels[0] - l0_length
            data_rep = torch.narrow(
                self.ten,
                dim=ops_dim,
                start=l0_length,
                length=self.ten.shape[ops_dim] - l0_length,
            )
            singleton_shape = tuple(
                data_rep.shape[d] if d == ops_dim else 1 for d in range(data_rep.dim())
            )
            # Subtract the L=0 offset in the pointer tensor
            idx_ten = (
                self.rep_layout[0][2, l0_length:]
                .sub(l0_length)
                .view(singleton_shape)
                .expand_as(data_rep)
            )
            invariant_rep = NormContraction1d.apply(
                data_rep, idx_ten, norm_shape, ops_dim, self._norm_eps
            )
            return torch.cat([data_l0, invariant_rep], dim=ops_dim)
        elif len(self.rep_dims) == 2:
            singleton_shape_0 = tuple(
                self.ten.shape[d] if d == self.rep_dims[0] else 1
                for d in range(self.ten.dim())
            )
            singleton_shape_1 = tuple(
                self.ten.shape[d] if d == self.rep_dims[1] else 1
                for d in range(self.ten.dim())
            )
            idx_ten_0 = (
                self.rep_layout[0][2, :]
                .unsqueeze(1)
                .view(singleton_shape_0)
                .expand_as(self.ten)
            )
            idx_ten_1 = (
                self.rep_layout[1][2, :]
                .unsqueeze(0)
                .view(singleton_shape_1)
                .expand_as(self.ten)
            )
            idx_tens = torch.stack([idx_ten_0, idx_ten_1], dim=0)
            norm_shape = list(self.shape)
            norm_shape[self.rep_dims[0]] = self.num_channels[0]
            norm_shape[self.rep_dims[1]] = self.num_channels[1]
            invariant2d = NormContraction2d.apply(
                self.ten, idx_tens, norm_shape, self.rep_dims, self._norm_eps
            )
            return invariant2d
        else:
            raise NotImplementedError

    def generate_rep_layout(self) -> Tuple[torch.LongTensor, ...]:
        if len(self.rep_dims) == 1:
            return (self.generate_rep_layout_1d_(self.metadata[0]),)
        elif len(self.rep_dims) == 2:
            rep_layout_0 = self.generate_rep_layout_1d_(self.metadata[0])
            rep_layout_1 = self.generate_rep_layout_1d_(self.metadata[1])
            return rep_layout_0, rep_layout_1
        else:
            raise NotImplementedError

    @staticmethod
    def generate_rep_layout_1d_(metadata1d) -> torch.LongTensor:
        n_irreps_per_l = torch.arange(start=0, end=metadata1d.size(0)) * 2 + 1
        end_channelids = torch.cumsum(metadata1d, dim=0)
        start_channelids = torch.cat([torch.LongTensor([0]), end_channelids[:-1]])
        dst_ls = torch.repeat_interleave(
            torch.arange(n_irreps_per_l.size(0)), n_irreps_per_l * metadata1d
        )
        dst_ms = torch.repeat_interleave(
            torch.cat(
                [
                    torch.arange(-l, l + 1)
                    for l in torch.arange(start=0, end=metadata1d.size(0))
                ]
            ),
            torch.repeat_interleave(metadata1d, n_irreps_per_l),
        )
        ns = torch.arange(metadata1d.sum())
        dst_ns = torch.cat(
            [
                ns[start_channelids[l] : end_channelids[l]].repeat(n_irreps)
                for l, n_irreps in enumerate(n_irreps_per_l)
            ]
        )
        rep_layout = torch.stack([dst_ls, dst_ms, dst_ns], dim=0).long()
        assert rep_layout.shape[1] == torch.sum(n_irreps_per_l * metadata1d)

        return rep_layout

    def to(self, device):
        # Do not transfer metadata to GPU, as index tracking should be more efficient
        # on CPUs
        self.ten = self.ten.to(device)
        self.rep_layout = (layout.to(device) for layout in self.rep_layout)
        return self



import torch

from torch_gauge.verlet_list import VerletList


# The i-j inversion should return a transposed matrix when the graph is dense
def test_verletlist_inversion_dense():
    data = torch.rand(17, 17, 8, 6)
    mask = torch.ones(17, 17, dtype=torch.bool)
    vl_test = VerletList().from_mask(mask, 17, 17, {}, {"test_edata": data})
    inversed = vl_test.to_src_first_view(vl_test.edata["test_edata"])
    assert torch.all(data.permute(1, 0, 2, 3).eq(inversed))


def test_verletlist_inversion_sparse():
    data = torch.rand(17, 17, 8, 6)
    mask_pre = torch.rand(17, 17)
    mask = (mask_pre + mask_pre.t()) > 0.5
    vl_test = VerletList().from_mask(mask, 32, 17, {}, {"test_edata": data})
    assert torch.all(
        data[mask, ...].eq(vl_test.edata["test_edata"][vl_test.edge_mask, ...])
    )

    inversed = vl_test.to_src_first_view(vl_test.edata["test_edata"])
    sparse1 = torch.zeros_like(data)
    sparse1[mask, ...] = data[mask, ...]
    sparse2 = torch.zeros_like(data)
    sparse2[mask, ...] = inversed[vl_test.edge_mask, ...]
    assert torch.all(sparse1.permute(1, 0, 2, 3).eq(sparse2))


# When the graph is dense, query_src inserts a dimension
def test_verletlist_querysrc_dense():
    data = torch.rand(17, 8, 6)
    mask = torch.ones(17, 17, dtype=torch.bool)
    vl_test = VerletList().from_mask(mask, 17, 17, {"test_ndata": data}, {})
    src_data = vl_test.query_src(vl_test.ndata["test_ndata"])
    assert torch.all(data.unsqueeze(0).expand(17, 17, 8, 6).eq(src_data))


if __name__ == "__main__":
    test_verletlist_inversion_dense()
    test_verletlist_querysrc_dense()



"""
Wigner (rotation) matrices for complex and real spherical harmonics
Evaluation and symbolic convention follows:
 Sakurai, J. J. "Modern Quantum Mechanics 2Nd Edition." Person New International edition (2014).
 Page 238, Eq. 3.9.33.
"""

import math

from joblib import Memory

memory = Memory(".o3_cache", verbose=0)
import torch
from scipy.special import factorial


@memory.cache
def _wigner_small_d_coeff(j, m1, m, k):
    sgn = (-1) ** (k - m + m1)
    numer = math.sqrt(
        factorial(j + m) * factorial(j - m) * factorial(j + m1) * factorial(j - m1)
    )
    denom = (
        factorial(j + m - k)
        * factorial(k)
        * factorial(j - k - m1)
        * factorial(k - m + m1)
    )
    if denom == 0:
        return None
    else:
        return sgn * numer / denom


def wigner_small_d_csh(j, beta):
    """Rotating complex spherical harmonics by the y axis"""
    small_d = torch.zeros(2 * j + 1, 2 * j + 1, dtype=torch.double)
    for m in range(-j, j + 1):
        for m1 in range(-j, j + 1):
            kmin = min(0, m - m1)
            kmax = max(0, j + m, j - m1)
            d_m1_m = 0
            for k in range(kmin, kmax + 1):
                prefactor = _wigner_small_d_coeff(j, m1, m, k)
                if prefactor is not None:
                    angular = (math.cos(beta / 2)) ** (2 * j - 2 * k + m - m1) * (
                        math.sin(beta / 2)
                    ) ** (2 * k - m + m1)
                    d_m1_m += prefactor * angular
            small_d[m1 + j, m + j] = d_m1_m
    return small_d


def wigner_D_csh(j, alpha, beta, gamma):
    """
    Euler rotation for complex spherical harmonics
    Contra-variant convention, for right multiplication
    """
    ms = torch.arange(-j, j + 1)
    small_d = wigner_small_d_csh(j, beta).to(torch.cdouble)
    alpha_phases = torch.exp(-1j * ms * alpha)
    gamma_phases = torch.exp(-1j * ms * gamma)
    wigner_D = alpha_phases.unsqueeze(1) * small_d * gamma_phases.unsqueeze(0)
    return wigner_D


@memory.cache
def csh_to_rsh(j):
    transform_mat = torch.zeros(2 * j + 1, 2 * j + 1, dtype=torch.cdouble)
    for m in range(-j, j + 1):
        if m < 0:
            transform_mat[j + m, j + m] = 1j / math.sqrt(2)
            transform_mat[j - m, j + m] = -((-1) ** m) * 1j / math.sqrt(2)
        elif m == 0:
            transform_mat[j + m, j + m] = 1
        elif m > 0:
            transform_mat[j - m, j + m] = 1 / math.sqrt(2)
            transform_mat[j + m, j + m] = (-1) ** m / math.sqrt(2)
        else:
            raise ValueError
    return transform_mat


def wigner_D_rsh(j, alpha, beta, gamma):
    c2r = csh_to_rsh(j)
    wigner_csh = wigner_D_csh(j, alpha, beta, gamma)
    wigner_rsh = (c2r.conj().t()).mm(wigner_csh).mm(c2r)
    # Checking the RSH rotation matrix entries are all real
    assert torch.allclose(
        wigner_rsh.imag, torch.zeros(2 * j + 1, 2 * j + 1, dtype=torch.double)
    ), print(c2r, wigner_csh, wigner_rsh)
    return wigner_rsh.real



"""
Geometric operations on point-clouds and manifolds
"""
import torch


class UnivecAngle(torch.autograd.Function):
    """
    Backward-robust angle calculation between two unit vectors
    See https://scicomp.stackexchange.com/questions/27689/numerically-stable-way-of-computing-angles-between-vectors
    Also see https://www.cs.utexas.edu/users/evouga/uploads/4/5/6/8/45689883/turning.pdf which is algebraically
    simpler but involves more forward arithmetic operations
    On the benchmark CPU, tensor.norm() is much slower than dot and 2x slower than
    cross product. We take the second approach here
    """

    @staticmethod
    def forward(ctx, vec1, vec2):
        cross_vec = torch.cross(vec1, vec2, dim=-1)
        cross_norm = cross_vec.pow(2).sum(dim=-1).sqrt()
        vec_dot = vec1.mul(vec2).sum(-1)
        angle = torch.atan2(cross_norm, 1 + vec_dot).mul_(2)
        ctx.save_for_backward(vec1, vec2, cross_vec, cross_norm)
        return angle

    @staticmethod
    def backward(ctx, grad_output):
        vec1, vec2, cross_vec, cross_norm = ctx.saved_tensors
        cross_norm = cross_norm.unsqueeze(-1).expand_as(cross_vec)
        z = cross_vec.div(cross_norm)
        # Need to consider using explicit Chebyshev-poly if this breaks in near-edge cases
        z.masked_fill_(cross_norm == 0, 0)
        grad_vec1, grad_vec2 = torch.cross(vec1, z), torch.cross(vec2, z).neg_()
        return grad_vec1.mul(grad_output.unsqueeze(-1)), grad_vec2.mul(
            grad_output.unsqueeze(-1)
        )


def univec_angle(vec1: torch.Tensor, vec2: torch.Tensor):
    return UnivecAngle.apply(vec1, vec2)


def univec_angle_unsafe(vec1: torch.Tensor, vec2: torch.Tensor):
    return torch.acos(vec1.mul(vec2).sum(-1))


def univec_cos(vec1: torch.Tensor, vec2: torch.Tensor):
    """Returns cos(\theta) for angular embeddings"""
    return vec1.mul(vec2).sum(-1)


class Chebyshev(nn.Module):
    """Generating Chebshev polynomials from powers"""

    def __init__(self, order):
        super(Chebyshev, self).__init__()
        self.order = order
        self.orders = torch.arange(order + 1, dtype=torch.long)
        self.orders = nn.Parameter(self.orders, requires_grad=False)
        self._init_coefficients()

    def _init_coefficients(self):
        """
        Generating combination coefficients using recursion
        Starting from order 1
        """
        coeff = torch.zeros(self.order + 1, self.order + 1)
        coeff[0, 0] = 1
        coeff[1, 1] = 1
        for idx in range(2, self.order + 1):
            coeff[1:, idx] = 2 * coeff[:-1, idx - 1]
            coeff[:, idx] -= coeff[:, idx - 2]
        self.cheb_coeff = nn.Parameter(coeff[:, 1:], requires_grad=False)

    def forward(self, x):
        size = (*x.shape, self.order + 1)
        x = x.unsqueeze(-1).expand(size).pow(self.orders)
        out = torch.matmul(x, self.cheb_coeff)
        return out



from dataclasses import dataclass
from typing import List

import torch

from torch_gauge.o3.spherical import SphericalTensor


@dataclass
class VerletList:
    def __init__(self):
        """
        Assuming in-edges and out-edges are identical (undirected graph)
        WARNING: the current Verlet-list scheme will break vanilla edge BatchNorm
        """
        self.neighbor_idx = None
        self.ndata = {}
        self.edata = {}
        self.edge_mask = None
        self.n_nodes = None
        self.PADSIZE = None
        self.batch_num_nodes = None
        self._dst_edim_locators = None

    def from_mask(
        self,
        verlet_mask: torch.BoolTensor,
        padding_size,
        num_nodes,
        one_body_data,
        two_body_data,
    ):
        self.PADSIZE = padding_size
        self.n_nodes = num_nodes
        self.batch_num_nodes = torch.LongTensor([self.n_nodes])

        in_degrees = verlet_mask.long().sum(1)
        src_raw = (
            torch.arange(num_nodes, dtype=torch.long)
            .unsqueeze(0)
            .expand(num_nodes, num_nodes)[verlet_mask]
        )

        # Scatter the src node-ids to the (N*PADDING_SIZE) padded table
        src_locators_1d = torch.LongTensor(
            [
                dstid * self.PADSIZE + src_location
                for dstid in range(self.n_nodes)
                for src_location in range(in_degrees[dstid])
            ]
        )
        src_edim_locators_flattened = torch.LongTensor(
            [
                src_location
                for dstid in range(self.n_nodes)
                for src_location in range(in_degrees[dstid])
            ]
        )
        src_edim_locators = torch.zeros(num_nodes, num_nodes, dtype=torch.long)
        src_edim_locators[verlet_mask] = src_edim_locators_flattened

        self.neighbor_idx = (
            torch.zeros(self.n_nodes * self.PADSIZE, dtype=torch.long)
            .scatter_(
                dim=0,
                index=src_locators_1d,
                src=src_raw,
            )
            .view(self.n_nodes, self.PADSIZE)
        )

        self.edge_mask = (
            torch.zeros(self.n_nodes * self.PADSIZE, dtype=torch.bool)
            .scatter_(
                dim=0,
                index=src_locators_1d,
                src=torch.ones(src_raw.size(0), dtype=torch.bool),
            )
            .view(self.n_nodes, self.PADSIZE)
        )

        self._dst_edim_locators = (
            torch.zeros(self.n_nodes, self.PADSIZE, dtype=torch.long)
            .masked_scatter_(
                mask=self.edge_mask,
                source=src_edim_locators.t()[verlet_mask],
            )
            .view(self.n_nodes, self.PADSIZE)
        )

        self.ndata = one_body_data
        # edata can be understood as node data with additional "neighborhood" dimensions
        # To make padding easier, always unsqueeze at least one trailing dimension
        self.edata = {
            k: self._scatter_efeat(edata, verlet_mask, src_raw)
            for k, edata in two_body_data.items()
        }

        return self

    def from_dgl(self, g: "dgl.DGLGraph", padding_size, nkeys, ekeys):
        """
        The interface for generating Verlet-list from dgl.DGLGraph
        Only scalar type tensors are supported
        """
        self.PADSIZE = padding_size
        self.n_nodes = g.num_nodes()
        self.batch_num_nodes = torch.LongTensor([self.n_nodes])

        src_raw, dst_raw, eid_raw = g.in_edges(torch.arange(self.n_nodes), form="all")
        in_degrees = g.in_degrees(torch.arange(self.n_nodes))

        # Scatter the src node-ids to the (N*PADDING_SIZE) padded table
        src_locators_1d = torch.LongTensor(
            [
                dstid * self.PADSIZE + src_location
                for dstid in range(self.n_nodes)
                for src_location in range(in_degrees[dstid])
            ]
        )
        src_edim_locators = torch.LongTensor(
            [
                src_location
                for dstid in range(self.n_nodes)
                for src_location in range(in_degrees[dstid])
            ]
        )
        g.edata["dst_edim_locators"] = torch.zeros_like(src_edim_locators)
        g.edata["dst_edim_locators"][g.edge_ids(src_raw, dst_raw)] = src_edim_locators
        self._dst_edim_locators = (
            torch.zeros(self.n_nodes * self.PADSIZE, dtype=src_raw.dtype)
            .scatter_(
                dim=0,
                index=src_locators_1d,
                src=g.edata["dst_edim_locators"][eid_raw],
            )
            .view(self.n_nodes, self.PADSIZE)
        )

        self.neighbor_idx = (
            torch.zeros(self.n_nodes * self.PADSIZE, dtype=src_raw.dtype)
            .scatter_(
                dim=0,
                index=src_locators_1d,
                src=src_raw,
            )
            .view(self.n_nodes, self.PADSIZE)
        )

        self.edge_mask = (
            torch.zeros(self.n_nodes * self.PADSIZE, dtype=torch.bool)
            .scatter_(
                dim=0,
                index=src_locators_1d,
                src=torch.ones(eid_raw.size(0), dtype=torch.bool),
            )
            .view(self.n_nodes, self.PADSIZE)
        )

        self.ndata = {k: g.ndata[k] for k in nkeys}
        self.edata = {
            k: torch.zeros(
                self.n_nodes * self.PADSIZE,
                g.edata[k][0, ...].numel(),
                dtype=g.edata[k].dtype,
            )
            .scatter_(
                dim=0,
                index=src_locators_1d.unsqueeze(1).expand(
                    -1, g.edata[k][0, ...].numel()
                ),
                src=g.edata[k][eid_raw, ...].view(g.edata[k].shape[0], -1),
            )
            .view(self.n_nodes, self.PADSIZE, *g.edata[k].shape[1:])
            for k in ekeys
        }
        return self

    def _scatter_efeat(self, edata, verlet_mask, src_raw):
        if isinstance(edata, torch.Tensor):
            return (
                torch.zeros(
                    self.n_nodes,
                    self.PADSIZE,
                    *edata.shape[2:],
                    dtype=edata.dtype,
                )
                .view(self.n_nodes, self.PADSIZE, -1)
                .masked_scatter_(
                    mask=self.edge_mask.unsqueeze(-1),
                    source=edata[verlet_mask, ...].view(src_raw.shape[0], -1),
                )
                .view(self.n_nodes, self.PADSIZE, *edata.shape[2:])
            )
        elif isinstance(edata, SphericalTensor):
            out_ten = (
                torch.zeros(
                    self.n_nodes,
                    self.PADSIZE,
                    *edata.shape[2:],
                    dtype=edata.ten.dtype,
                )
                .view(self.n_nodes, self.PADSIZE, -1)
                .masked_scatter_(
                    mask=self.edge_mask.unsqueeze(-1),
                    source=edata.ten[verlet_mask, ...].view(src_raw.shape[0], -1),
                )
                .view(self.n_nodes, self.PADSIZE, *edata.shape[2:])
            )
            return SphericalTensor(
                out_ten,
                rep_dims=edata.rep_dims,
                metadata=edata.metadata,
                rep_layout=edata.rep_layout,
                num_channels=edata.num_channels,
            )

    def query_src(self, src_feat):
        """
        Returns the src-node data scattered into the neighbor-list frame.
        When applied to an edge-data tensor, this function generates a higher-order
         view (k+1 hop) of the underlying k-hop graph structure.
        src_feat must be contiguous.
        """
        flattened_neighboridx = self.neighbor_idx.view(-1)
        if isinstance(src_feat, torch.Tensor):
            flattened2d_src = src_feat.view(src_feat.shape[0], -1)
            flattened_out = (
                flattened2d_src[flattened_neighboridx, ...]
                .view(*self.neighbor_idx.shape, flattened2d_src.shape[1])
                .mul_(self.edge_mask.unsqueeze(2))
            )
            return flattened_out.view(*self.neighbor_idx.shape, *src_feat.shape[1:])
        elif isinstance(src_feat, SphericalTensor):
            flattened2d_src = src_feat.ten.view(src_feat.ten.shape[0], -1)
            flattened_out = (
                flattened2d_src[flattened_neighboridx, ...]
                .view(*self.neighbor_idx.shape, flattened2d_src.shape[1])
                .mul_(self.edge_mask.unsqueeze(2))
            )
            return SphericalTensor(
                flattened_out.view(*self.neighbor_idx.shape, *src_feat.shape[1:]),
                rep_dims=tuple(dim + 1 for dim in src_feat.rep_dims),
                metadata=src_feat.metadata,
                rep_layout=src_feat.rep_layout,
                num_channels=src_feat.num_channels,
            )

    def to_src_first_view(self, data):
        """
        Flipping src / dst node indexing order without inverting the data
        """
        scatter_ten = (self.neighbor_idx * self.PADSIZE + self._dst_edim_locators)[
            self.edge_mask
        ]
        if isinstance(data, torch.Tensor):
            out_ten = torch.zeros_like(data).view(self.n_nodes * self.PADSIZE, -1)
            out_ten[scatter_ten, :] = data.view(self.n_nodes, self.PADSIZE, -1)[
                self.edge_mask, :
            ]
            return out_ten.view_as(data)
        elif isinstance(data, SphericalTensor):
            out_ten = torch.zeros_like(data.ten).view(self.n_nodes * self.PADSIZE, -1)
            out_ten[scatter_ten, :] = data.ten.view(self.n_nodes, self.PADSIZE, -1)[
                self.edge_mask, :
            ]
            return SphericalTensor(
                out_ten.view_as(data.ten),
                rep_dims=data.rep_dims,
                metadata=data.metadata,
                rep_layout=data.rep_layout,
                num_channels=data.num_channels,
            )
        else:
            raise NotImplementedError

    def to(self, device):
        self.neighbor_idx = self.neighbor_idx.to(device)
        self.ndata = {k: v.to(device) for k, v in self.ndata.items()}
        self.edata = {k: v.to(device) for k, v in self.edata.items()}
        self.edge_mask = self.edge_mask.to(device)
        self._dst_edim_locators = self._dst_edim_locators.to(device)
        self.batch_num_nodes = self.batch_num_nodes.to(device)
        return self

    @staticmethod
    def batch(vls: List["VerletList"]):
        """
        WARNING: In the current version, taking batch of batches will
         break the offset indices
        """
        batched_vl = VerletList()
        batched_vl.PADSIZE = vls[0].PADSIZE
        batched_vl.batch_num_nodes = torch.cat([vl.batch_num_nodes for vl in vls])
        batched_vl.n_nodes = torch.sum(batched_vl.batch_num_nodes)
        bnn_offsets = torch.repeat_interleave(
            torch.cat(
                [torch.LongTensor([0]), torch.cumsum(batched_vl.batch_num_nodes, dim=0)]
            )[:-1],
            batched_vl.batch_num_nodes,
        )
        batched_vl.neighbor_idx = torch.cat(
            [vl.neighbor_idx for vl in vls], dim=0
        ) + bnn_offsets.unsqueeze(1)
        batched_vl.edge_mask = torch.cat([vl.edge_mask for vl in vls], dim=0)
        batched_vl.ndata = {
            nk: torch.cat([vl.ndata[nk] for vl in vls], dim=0)
            for nk in vls[0].ndata.keys()
        }
        batched_vl.edata = {
            ek: torch.cat([vl.edata[ek] for vl in vls], dim=0)
            for ek in vls[0].edata.keys()
        }
        batched_vl._dst_edim_locators = torch.cat(
            [vl._dst_edim_locators for vl in vls], dim=0
        )

        return batched_vl



from setuptools import setup, find_packages
import versioneer

setup(
    name="torch_gauge",
    description="A light-weight PyTorch extension for gauge-equivariant geometric learning",
    author="Zhuoran Qiao",
    author_email="zqiao@caltech.edu",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_packages(),
    python_requires=">=3.7",
    install_requires=['torch>=1.7.0'],
)



"""
Testing real spherical harmonics generation
"""

import math
import random

import torch

from torch_gauge.o3.rsh import RSHxyz
from torch_gauge.o3.wigner import wigner_D_rsh


def test_batch_rsh_generation():
    rshmodule = RSHxyz(max_l=6)
    xyz = torch.normal(0, 1, size=(1024, 3))
    rshs = rshmodule(xyz)
    assert torch.all(rshs.metadata.eq(torch.tensor([[1, 1, 1, 1, 1, 1, 1]])))


def test_rsh_explicit():
    rshmodule = RSHxyz(max_l=2)
    # Compare with the explicit formula up to l=2
    for _ in range(100):
        x = torch.normal(0, 1, size=(1,)).double()
        y = torch.normal(0, 1, size=(1,)).double()
        z = torch.normal(0, 1, size=(1,)).double()
        r = torch.sqrt(x ** 2 + y ** 2 + z ** 2)
        rshs_ref = torch.tensor(
            [
                1,
                y,
                z,
                x,
                math.sqrt(3) * x * y,
                math.sqrt(3) * y * z,
                1 / 2 * (3 * z ** 2 - r ** 2),
                math.sqrt(3) * x * z,
                (math.sqrt(3) / 2) * (x ** 2 - y ** 2),
            ]
        )
        assert torch.allclose(
            rshmodule(torch.stack([x, y, z], dim=1)).ten,
            rshs_ref.unsqueeze(0),
            atol=1e-8,
            rtol=0,
        )


def test_rsh_batch_explicit():
    rshmodule = RSHxyz(max_l=2)
    x = torch.normal(0, 1, size=(1024,)).double()
    y = torch.normal(0, 1, size=(1024,)).double()
    z = torch.normal(0, 1, size=(1024,)).double()
    r = torch.sqrt(x ** 2 + y ** 2 + z ** 2)
    rshs_ref = torch.stack(
        [
            torch.ones_like(x),
            y,
            z,
            x,
            math.sqrt(3) * x * y,
            math.sqrt(3) * y * z,
            1 / 2 * (3 * z ** 2 - r ** 2),
            math.sqrt(3) * x * z,
            (math.sqrt(3) / 2) * (x ** 2 - y ** 2),
        ],
        dim=1,
    )
    assert torch.allclose(
        rshmodule(torch.stack([x, y, z], dim=1)).ten,
        rshs_ref,
        atol=1e-8,
        rtol=0,
    )


def test_wigner_rsh_z():
    for _ in range(10):
        alpha = random.random() * 2 * math.pi
        rsh_wigner_1 = wigner_D_rsh(1, alpha, 0, 0)
        ref_rotmat = torch.tensor(
            [
                [math.cos(alpha), 0, -math.sin(alpha)],
                [0, 1, 0],
                [math.sin(alpha), 0, math.cos(alpha)],
            ],
            dtype=torch.double,
        ).t()
        assert torch.allclose(rsh_wigner_1, ref_rotmat, atol=1e-7, rtol=1e-5,), print(
            "Wigner-D at j=1 does not match the Cartesian rotation matrix: ",
            rsh_wigner_1,
            ref_rotmat,
        )



"""
Native implementation of point-cloud based graph neural networks
 with torch-gauge functionalities
"""


import torch
from torch.nn import Linear, Parameter

from torch_gauge.nn import SSP
from torch_gauge.verlet_list import VerletList


class SchNetLayer(torch.nn.Module):
    """
    Schütt, Kristof, et al. "Schnet: A continuous-filter convolutional neural network for modeling quantum interactions."
     Advances in neural information processing systems. 2017.
    """

    def __init__(self, num_features):
        super().__init__()
        _nf = num_features
        self.gamma = 10.0
        self.rbf_centers = Parameter(
            torch.linspace(0.1, 30.1, 300), requires_grad=False
        )
        self.cfconv = torch.nn.Sequential(
            Linear(300, _nf), SSP(), Linear(_nf, _nf), SSP()
        )
        self.pre_conv = Linear(_nf, _nf)
        self.post_conv = torch.nn.Sequential(Linear(_nf, _nf), SSP(), Linear(_nf, _nf))

    def forward(self, vl: VerletList, l: int):
        xyz = vl.ndata["xyz"]
        pre_conv = self.pre_conv(vl.ndata[f"atomic_{l}"])
        d_ij = (vl.query_src(xyz) - xyz.unsqueeze(1)).norm(dim=2, keepdim=True)
        filters = self.cfconv(
            torch.exp(-self.gamma * (d_ij - self.rbf_centers.view(1, 1, -1)).pow(2))
        )
        conv_out = (filters * vl.query_src(pre_conv) * vl.edge_mask.unsqueeze(2)).sum(1)
        vl.ndata[f"atomic_{l+1}"] = vl.ndata[f"atomic_{l}"] + self.post_conv(conv_out)
        return vl.ndata[f"atomic_{l+1}"]
